import csv

from duckietown_world import SE2Transform
from duckietown_world.world_duckietown.tags_db import get_sign_type_from_tag_id

in_file = open("autolab_tiles_map.csv", "rb")
reader = csv.reader(in_file)

data = list(reader)[1:]

map_data = {}
map_data["version"] = 2

H = W = 0
for entry in data:
    xs, ys, kind, rotation = entry
    x = int(xs)
    y = int(ys)
    H = max(H, y + 1)
    W = max(W, x + 1)

tiles = [[None for _ in range(W)] for __ in range(H)]

for entry in data:
    xs, ys, kind, rotation = entry
    x = int(xs)
    y = int(ys)
    rotation = int(rotation)
    direction = {90: "N", 180: "W", 270: "S", 0: "E"}
    translate = {"empty": "asphalt", "turn": "curve_left", "3way": "3way_left"}
    kind = translate.get(kind, kind)
    k = kind + "/" + direction[rotation]
    tiles[H - 1 - y][x] = k
map_data["tiles"] = tiles
#
# tiles2 = []
# for row in tiles:
#     row_string = '  '.join('%10s' % _ for _ in row)
#     tiles2.append(row_string)
#
#
#
# map_data['tiles-string'] = "\n".join(tiles2)

import yaml

in_file = open("autolab_tags_map.csv", "rb")
reader = csv.reader(in_file)
data = list(reader)[1:]

objects = map_data["objects"] = {}
import numpy as np

APRIL_TAG_SIZE = 0.08

for i, entry in enumerate(data):
    tag_ID, x, y, slot, rotation = entry
    tag_ID = int(tag_ID)

    kind = get_sign_type_from_tag_id(tag_ID)
    x = int(x)
    y = int(y)
    slot = int(slot)

    rotation = rotation or 0
    rotation = int(rotation)
    # bug in the map file
    # rotation = rotation + 90
    sign = dict(
        kind=kind,
        tag={"~TagInstance": dict(tag_id=int(tag_ID), family="36h11", size=APRIL_TAG_SIZE)},
        attach=dict(tile=[x, y], slot=slot),
    )

    sign["pose"] = SE2Transform([0, 0], np.deg2rad(rotation)).as_json_dict()

    objects["tag%02d" % i] = sign

in_file = open("AprilTag position duckietown.csv", "rb")
reader = csv.reader(in_file)
data = list(reader)[1:]
CM = 0.01
origin = -5 * CM, -15.5 * CM
# Tag ID,Quadrant,x location,Y-location,Roation

# FAMILY = 'default'
# SIZE = 0.035
for entry in data:
    tag_ID, quadrant, x, y, rotation = entry
    x = float(x) + origin[0]
    y = float(y) + origin[1]

    rotation = rotation or 0
    rotation = int(rotation)

    sign = dict(
        kind="floor_tag", tag={"~TagInstance": dict(tag_id=int(tag_ID), family="36h11", size=APRIL_TAG_SIZE)},
    )
    sign["pose"] = SE2Transform([x, y], np.deg2rad(rotation)).as_json_dict()

    objects[quadrant] = sign

s = yaml.safe_dump(map_data, encoding="utf-8", indent=4, allow_unicode=True, default_flow_style=False)
print(s)



# coding=utf-8
import os

import numpy as np
from comptests import comptest, run_module_tests, get_comptests_output_dir

import geometry as geo
from duckietown_world import SE2Transform, RectangularArea, list_maps
from duckietown_world.svg_drawing import draw_static
from duckietown_world.svg_drawing.draw_maps import draw_map
from duckietown_world.world_duckietown.lane_segment import LaneSegment
from duckietown_world.world_duckietown.map_loading import load_map
from duckietown_world.world_duckietown.tile_template import load_tile_types


@comptest
def svg1():
    outdir = get_comptests_output_dir()
    control_points = [
        SE2Transform.from_SE2(geo.SE2_from_translation_angle([0, 0], 0)),
        SE2Transform.from_SE2(geo.SE2_from_translation_angle([1, 0], 0)),
        SE2Transform.from_SE2(geo.SE2_from_translation_angle([2, -1], np.deg2rad(-90))),
    ]

    width = 0.3

    ls = LaneSegment(width, control_points)

    area = RectangularArea((-1, -2), (3, 2))
    draw_static(ls, outdir, area=area)


@comptest
def svg2():
    outdir = get_comptests_output_dir()
    templates = load_tile_types()

    for k, v in templates.items():
        area = RectangularArea((-1, -1), (1, 1))
        dest = os.path.join(outdir, k)
        draw_static(v, dest, area=area)


@comptest
def maps():
    outdir = get_comptests_output_dir()

    map_names = list_maps()
    print(map_names)

    for map_name in map_names:
        duckietown_map = load_map(map_name)
        out = os.path.join(outdir, map_name)
        draw_map(out, duckietown_map)


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8

import copy
from dataclasses import dataclass, field
from typing import Dict, List, Tuple, Union, Callable

import yaml
from duckietown_serialization_ds1 import Serializable


from duckietown_world.seqs import UndefinedAtTime
from zuper_commons.text import indent
from zuper_commons.types import check_isinstance
from .rectangular_area import RectangularArea
from .transforms import Transform

__all__ = ["PlacedObject", "SpatialRelation", "GroundTruth", "get_object_tree", "FQN"]

FQN = Tuple[str, ...]


@dataclass
class SpatialRelation(Serializable):
    a: Tuple[str]
    transform: Transform
    b: Tuple[str]

    def filter_all(self, f):
        t2 = f(self.transform)
        if t2 == self.transform:
            return self
        else:
            return SpatialRelation(self.a, t2, self.b)

    @classmethod
    def params_from_json_dict(cls, d):
        a = d.pop("a", [])
        b = d.pop("b")
        transform = d.pop("transform")
        transform = Serializable.from_json_dict(transform)
        return dict(a=a, b=b, transform=transform)

    def params_to_json_dict(self):
        res = {}
        if self.a:
            res["a"] = list(self.a)
        res["b"] = self.b
        res["transform"] = self.transform
        return res


class GroundTruth(SpatialRelation):
    def __repr__(self):
        return "GroundTruth(%r -> %r  %s)" % (self.a, self.b, self.transform)

    @classmethod
    def params_from_json_dict(cls, d):
        return {}

    def params_to_json_dict(self):
        return {}


root: FQN = ()


@dataclass
class PlacedObject(Serializable):
    children: Dict[str, "PlacedObject"] = field(default_factory=dict)
    spatial_relations: Dict[str, SpatialRelation] = field(default_factory=dict)

    def __post_init__(self):
        from .transforms import Transform

        for k, v in list(self.spatial_relations.items()):
            if isinstance(v, Transform):
                if k in self.children:
                    b: Tuple[str] = (k,)
                    sr = GroundTruth(a=root, b=b, transform=v)
                    self.spatial_relations[k] = sr
                else:
                    msg = 'What is the "%s" referring to?' % k
                    raise ValueError(msg)

        if not self.spatial_relations:
            for child in self.children:
                from duckietown_world import SE2Transform

                sr = GroundTruth(a=root, b=(child,), transform=SE2Transform.identity())
                self.spatial_relations[child] = sr

    def remove_object(self, k: str) -> None:
        self.children.pop(k)
        for sr_id, sr in list(self.spatial_relations.items()):
            if sr.b == (k,):
                self.spatial_relations.pop(sr_id)

    def _simplecopy(self, *args, **kwargs):
        children = dict((k, v) for k, v in self.children.items())
        spatial_relations = dict((k, v) for k, v in self.spatial_relations.items())
        kwargs.update(dict(children=children, spatial_relations=spatial_relations))
        return type(self)(*args, **kwargs)

    def _copy(self):
        if type(self) is PlacedObject:
            return self._simplecopy()
        else:
            # logger.debug('no _copy for %s' % type(self).__name__)
            return copy.copy(self)

    def filter_all(self, f: "Callable[[PlacedObject], PlacedObject]") -> "PlacedObject":
        children = {}
        spatial_relations = {}

        no_child = []
        for child_name, child in list(self.children.items()):
            try:
                child2 = f(child.filter_all(f))
            except UndefinedAtTime:
                no_child.append(child_name)
            else:
                children[child_name] = child2

        for sr_name, sr in list(self.spatial_relations.items()):
            if sr.b and sr.b[0] in no_child:
                pass
            else:
                try:
                    sr2 = f(sr.filter_all(f))
                except UndefinedAtTime:
                    pass
                else:
                    spatial_relations[sr_name] = sr2

        if children == self.children and spatial_relations == self.spatial_relations:
            # logger.debug('could save a copy of %s' % type(self).__name__)
            return self
        else:
            # print('need a copy of %s' % self)
            x = self._copy()
            x.children = children
            x.spatial_relations = spatial_relations
            return x

    def __getitem__(self, item: Union[str, FQN]) -> "PlacedObject":
        """

        Either url-like:

            child1/sub
            .

        or tuple like:

            ('child1', 'sub')
            ()

        :param item:
        :return:
        """
        if isinstance(item, str):
            item = fqn_from_url(item)
        check_isinstance(item, tuple)
        return self.get_object_from_fqn(item)

    def get_object_from_fqn(self, fqn: FQN) -> "PlacedObject":
        if fqn == ():
            return self
        first, rest = fqn[0], fqn[1:]
        if first in self.children:
            return self.children[first].get_object_from_fqn(rest)
        else:
            msg = "Cannot find child %s in %s" % (first, list(self.children))
            raise KeyError(msg)

    def params_to_json_dict(self):
        res = {}

        if self.children:
            res["children"] = self.children
        if self.spatial_relations:
            res["spatial_relations"] = self.spatial_relations

        return res

    def set_object(self, name: str, ob: "PlacedObject", **transforms: SpatialRelation):
        assert self is not ob
        self.children[name] = ob
        type2klass = {"ground_truth": GroundTruth}
        root: Tuple[str, ...] = ()
        for k, v in transforms.items():
            klass = type2klass[k]
            # noinspection PyArgumentList
            st = klass(a=root, b=(name,), transform=v)
            i = len(self.spatial_relations)
            self.spatial_relations[i] = st

    def draw_svg(self, drawing, g):
        from duckietown_world.svg_drawing import draw_axes

        draw_axes(drawing, g)

    def get_drawing_children(self) -> List[str]:
        return sorted(self.children)

    def extent_points(self):
        return [(0.0, 0.1), (0.1, 0.0)]

    def get_footprint(self):
        return RectangularArea([-0.1, -0.1], [0.1, 0.1])


def get_child_transform(po: PlacedObject, child: str) -> Transform:
    for _ in po.spatial_relations.values():
        if _.a == () and _.b == (child,):
            return _.transform
    raise KeyError(child)


def get_object_tree(
    po: PlacedObject, levels: int = 100, spatial_relations: bool = False, attributes: bool = False,
) -> str:
    ss = []
    ss.append("%s" % type(po).__name__)
    d = po.params_to_json_dict()
    d.pop("children", None)
    d.pop("spatial_relations", None)

    if attributes:
        if d:
            ds = yaml.safe_dump(d, encoding="utf-8", indent=4, allow_unicode=True, default_flow_style=False,)
            if isinstance(ds, bytes):
                ds = ds.decode("utf-8")
            ss.append("\n" + indent(ds, " "))

    if po.children and levels >= 1:
        ss.append("")
        N = len(po.children)
        for i, (child_name, child) in enumerate(po.children.items()):

            if i != N - 1:
                prefix1 = u"├ %s ┐ " % child_name
                prefix2 = u"│ %s │ " % (" " * len(child_name))
            else:
                prefix1 = u"└ %s ┐ " % child_name
                prefix2 = u"  %s │ " % (" " * len(child_name))
            c = get_object_tree(
                child, attributes=attributes, spatial_relations=spatial_relations, levels=levels - 1,
            )
            sc = indent(c, prefix2, prefix1)
            n = max(len(_) for _ in sc.split("\n"))
            sc += "\n" + prefix2[:-2] + u"└" + u"─" * (n - len(prefix2) + 3)
            ss.append(sc)

    if spatial_relations:
        if po.spatial_relations and levels >= 1:
            ss.append("")
            for r_name, rel in po.spatial_relations.items():
                ss.append(
                    '- from "%s" to "%s"  %s ' % (url_from_fqn(rel.a), url_from_fqn(rel.b), rel.transform)
                )

    return "\n".join(ss)


def url_from_fqn(x):
    if not x:
        return "."
    else:
        return "/".join(x)


def fqn_from_url(u):
    if u == ".":
        return ()
    else:
        return tuple(u.split("/"))



# coding=utf-8
from dataclasses import dataclass

import numpy as np

import geometry as geo
from duckietown_world.world_duckietown.dynamics_delay import ApplyDelay
from .generic_kinematics import GenericKinematicsSE2
from .platform_dynamics import PlatformDynamicsFactory

# from duckietown_serialization_ds1 import Serializable

__all__ = ["DynamicModelParameters", "DynamicModel", "PWMCommands"]


@dataclass
class PWMCommands:
    """
        PWM commands are floats between -1 and 1.
    """

    motor_left: float
    motor_right: float


class DynamicModelParameters(PlatformDynamicsFactory):
    def __init__(self, u1, u2, u3, w1, w2, w3, uar, ual, war, wal):
        # parameters for autonomous dynamics
        self.u1 = u1
        self.u2 = u2
        self.u3 = u3
        self.w1 = w1
        self.w2 = w2
        self.w3 = w3

        # parameters for forced dynamics
        self.u_alpha_r = uar
        self.u_alpha_l = ual
        self.w_alpha_r = war
        self.w_alpha_l = wal

    def initialize(self, c0, t0=0, seed=None) -> "DynamicModel" "":
        return DynamicModel(self, c0, t0)


def get_DB18_nominal(delay: float) -> PlatformDynamicsFactory:
    # parameters for autonomous dynamics
    u1 = 5
    u2 = 0
    u3 = 0
    w1 = 4
    w2 = 0
    w3 = 0
    # parameters for forced dynamics
    uar = 1.5
    ual = 1.5
    war = 15  # modify this for trim
    wal = 15

    parameters = DynamicModelParameters(
        u1=u1, u2=u2, u3=u3, w1=w1, w2=w2, w3=w3, uar=uar, ual=ual, war=war, wal=wal
    )

    if delay > 0:
        delayed = ApplyDelay(parameters, delay, PWMCommands(0, 0))
        return delayed
    else:
        return parameters


def get_DB18_uncalibrated(delay: float, trim: float = 0) -> PlatformDynamicsFactory:
    # parameters for autonomous dynamics
    u1 = 5
    u2 = 0
    u3 = 0
    w1 = 4
    w2 = 0
    w3 = 0
    # parameters for forced dynamics
    uar = 1.5
    ual = 1.5
    war = 15 * (1.0 + trim)
    wal = 15 * (1.0 - trim)

    parameters = DynamicModelParameters(
        u1=u1, u2=u2, u3=u3, w1=w1, w2=w2, w3=w3, uar=uar, ual=ual, war=war, wal=wal
    )

    if delay > 0:
        delayed = ApplyDelay(parameters, delay, PWMCommands(0, 0))
        return delayed
    else:
        return parameters


class DynamicModel(GenericKinematicsSE2):
    """
        This represents a dynamical formulation of of a differential-drive vehicle.
    """

    parameters: DynamicModelParameters

    def __init__(self, parameters: DynamicModelParameters, c0, t0):
        self.parameters = parameters
        GenericKinematicsSE2.__init__(self, c0, t0)

    @staticmethod
    def model(input: PWMCommands, parameters: DynamicModelParameters, u=None, w=None):
        ## Unpack Inputs
        U = np.array([input.motor_right, input.motor_left])
        V = U.reshape(U.size, 1)
        V = np.clip(V, -1, +1)
        # parameters for autonomous dynamics
        u1 = parameters.u1
        u2 = parameters.u2
        u3 = parameters.u3
        w1 = parameters.w1
        w2 = parameters.w2
        w3 = parameters.w3
        # parameters for forced dynamics
        u_alpha_r = parameters.u_alpha_r
        u_alpha_l = parameters.u_alpha_l
        w_alpha_r = parameters.w_alpha_r
        w_alpha_l = parameters.w_alpha_l

        ## Calculate Dynamics
        # nonlinear Dynamics - autonomous response
        f_dynamic = np.array([[-u1 * u - u2 * w + u3 * w ** 2], [-w1 * w - w2 * u - w3 * u * w]])
        # input Matrix
        B = np.array([[u_alpha_r, u_alpha_l], [w_alpha_r, -w_alpha_l]])
        # forced response
        f_forced = np.matmul(B, V)
        # acceleration
        x_dot_dot = f_dynamic + f_forced

        return x_dot_dot

    def integrate(self, dt: float, commands: PWMCommands) -> "DynamicModel":
        # previous velocities (v0)
        linear_angular_prev = geo.linear_angular_from_se2(self.v0)
        linear_prev = linear_angular_prev[0]
        longit_prev = linear_prev[0]
        lateral_prev = linear_prev[1]
        angular_prev = linear_angular_prev[1]

        # predict the acceleration of the vehicle
        x_dot_dot = self.model(commands, self.parameters, u=longit_prev, w=angular_prev)

        # convert the acceleration to velocity by forward euler
        longitudinal = longit_prev + dt * x_dot_dot[0]
        angular = angular_prev + dt * x_dot_dot[1]
        lateral = 0.0

        linear = [longitudinal, lateral]

        # represent this as se(2)
        commands_se2 = geo.se2_from_linear_angular(linear, angular)

        # call the "integrate" function of GenericKinematicsSE2
        s1 = GenericKinematicsSE2.integrate(self, dt, commands_se2)

        # new state
        c1 = s1.q0, s1.v0
        t1 = s1.t0

        return DynamicModel(self.parameters, c1, t1)



# coding=utf-8
from abc import ABCMeta, abstractmethod

from .types import *

__all__ = ["PlatformDynamicsFactory", "PlatformDynamics"]


class PlatformDynamics(metaclass=ABCMeta):
    """
        This class represents the state of a dynamical system.


            s0 = ...

            s1 = s0.integrate(dt=0.1, commands=[1,1])
            s2 = s0.integrate(dt=0.1, commands=[1,1])


        Each subclass has its own representation of commands.

    """

    @abstractmethod
    def integrate(self, dt: float, commands):
        """
            Returns the result of applying commands for dt.

            :param dt > 0: time interval
            :param commands: class-specific commands
            :return: the next state
        """

    @abstractmethod
    def TSE2_from_state(self) -> TSE2v:
        """ Returns pose, velocity for the state. """


class PlatformDynamicsFactory(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def initialize(cls, c0: TSE2v, t0: float = 0, seed: int = None) -> PlatformDynamics:
        """
            Returns the dynamics initalized at a certain configuration.

            :param c0: configuration in TSE2
            :param t0: time in which to sstart
            :param seed: seed for a possible random number generator

        """



# coding=utf-8
from dataclasses import dataclass
from typing import *

from duckietown_serialization_ds1.serialization1 import as_json_dict

from .tsequence import GenericSequence, X, Timestamp

__all__ = ["Constant"]


@dataclass
class Constant(GenericSequence[X]):
    always: X

    def at(self, t: Timestamp) -> X:
        return self.always

    # noinspection PyUnusedLocal
    def at_or_previous(self, t: Timestamp) -> X:
        return self.always

    def get_sampling_points(self) -> str:
        return GenericSequence.CONTINUOUS

    def get_start(self) -> Optional[Timestamp]:
        return None

    def get_end(self) -> Optional[Timestamp]:
        return None

    def as_json_dict(self):
        return {"always": as_json_dict(self.always)}



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""

Heavily modified by Andrea Censi for PyMCDP.


"""

from sys import stdout
from six import StringIO

gvgen_version = "2.0"

debug = 0
debug_tree_unroll = 0


class ACGvGen(object):
    """
    Graphviz dot language Generation Class
    For example of usage, please see the __main__ function
    """

    def __init__(self, legend_name=None, options="compound=true;"):  # allow links between clusters
        self.max_line_width = 10
        self.max_arrow_width = 2
        self.line_factor = 1
        self.arrow_factor = 0.5
        self.initial_line_width = 1.2
        self.initial_arrow_width = 0.8

        self.options = {}
        for option in options.split(";"):
            option = option.strip()
            if not option:
                continue
            key, value = option.split("=", 1)
            self.setOptions(**{key: value})

        self.__id = 0
        self.__nodes = []
        self.__links = []
        self.__browse_level = 0  # Stupid depth level for self.browse
        self.__opened_braces = []  # We count opened clusters
        self.fd = stdout  # File descriptor to output dot
        self.padding_str = "   "  # Left padding to make children and parent look nice
        self.__styles = {}
        self.__default_style = []
        self.__default_style_links = []
        #         self.smart_mode = 0                      # Disabled by default

        # The graph has a legend
        if legend_name:
            self.setOptions(rankdir="LR")
            self.legend = self.newItem(legend_name)

    def setOptions(self, **options):
        for key, value in options.items():
            self.options[key] = value

    def __node_new(self, name, parent=None, distinct=None):
        """
        Create a new node in the data structure
        @name: Name of the node, that will be the graphviz label
        @parent: The node parent
        @distinct: if true, will not create and node that has the same name

        Returns: The node created
        """

        # We first check for distincts
        if distinct:
            if self.__nodes:
                for e in self.__nodes:
                    props = e["properties"]
                    if props["label"] == name:
                        # We found the label name matching, we return -1
                        return -1

        # We now insert into gvgen datastructure
        self.__id += 1
        node = {
            "id": self.__id,  # Internal ID
            "lock": 0,  # When the node is written, it is locked to avoid further references
            "parent": parent,  # Node parent for easy graphviz clusters
            "style": None,  # Style that GvGen allow you to create
            "properties": {  # Custom graphviz properties you can add, which will overide previously defined styles
                "label": name
            },
        }

        # Parents should be sorted first
        if parent:
            self.__nodes.insert(1, node)
        else:
            self.__nodes.append(node)

        return node

    #     def __link_smart(self, link):
    #         """
    #         Creates a smart link if smart_mode activated:
    #           if a -> b exists, and we now add a <- b,
    #           instead of doing:  a -> b
    #                                <-
    #           we do: a <-> b
    #         """

    # linkfrom = self.__link_exists(link['from_node'], link['to_node'])
    # linkto = self.__link_exists(link['to_node'], link['from_node'])

    #         if self.smart_mode:
    #             if linkto:
    #                 self.__links.remove(linkto)
    #                 self.propertyAppend(link, "dir", "both")
    #
    #             pw = self.propertyGet(linkfrom, "penwidth")
    #             if pw:
    #                 pw = float(pw)
    #                 pw += self.line_factor
    #                 if pw < self.max_line_width:
    #                     self.propertyAppend(linkfrom, "penwidth", str(pw))
    #             else:
    #                 self.propertyAppend(link, "penwidth", str(self.initial_line_width))
    #
    #             aw = self.propertyGet(linkfrom, "arrowsize")
    #             if aw:
    #                 aw = float(aw)
    #                 if aw < self.max_arrow_width:
    #                     aw += self.arrow_factor
    #                     self.propertyAppend(linkfrom, "arrowsize", str(aw))
    #             else:
    #                 self.propertyAppend(link, "arrowsize", str(self.initial_arrow_width))

    # if not linkfrom:
    #         self.__links.append(link)

    def __link_new(self, from_node, to_node, label=None, cl_from_node=None, cl_to_node=None):
        """
        Creates a link between two nodes
        @from_node: The node the link comes from
        @to_node: The node the link goes to

        Returns: The link created
        """

        link = {
            "from_node": from_node,
            "to_node": to_node,
            "style": None,  # Style that GvGen allow you to create
            "properties": {},
            # Custom graphviz properties you can add, which will overide previously defined styles
            "cl_from_node": None,  # When linking from a cluster, the link appears from this node
            "cl_to_node": None,  # When linking to a cluster, the link appears to go to this node
        }

        if label:
            link["properties"]["label"] = label

        if cl_from_node:
            link["cl_from_node"] = cl_from_node
        if cl_to_node:
            link["cl_to_node"] = cl_to_node

        # We let smart link work for us
        # self.__link_smart(link)

        self.__links.append(link)

        return link

    #     def __link_exists(self, from_node, to_node):
    #         """
    #         Find if a link exists
    #         @from_node: The node the link comes from
    #         @to_node: The node the link goes to
    #
    #         Returns: true if the given link already exists
    #         """
    #
    #         for link in self.__links:
    #             if link['from_node'] == from_node and link['to_node'] == to_node:
    #                 return link
    #
    #         return None

    def __has_children(self, parent):
        """
        Find children to a given parent
        Returns the children list
        """
        children_list = []
        for e in self.__nodes:
            if e["parent"] == parent:
                children_list.append(e)

        return children_list

    def newItem(self, name, parent=None, distinct=None):
        node = self.__node_new(name, parent, distinct)

        return node

    def newLink(self, src, dst, label=None, cl_src=None, cl_dst=None):
        """
        Link two existing nodes with each other
        """

        return self.__link_new(src, dst, label, cl_src, cl_dst)

    def debug(self):
        for e in self.__nodes:
            print("element = " + str(e["id"]))

    def collectLeaves(self, parent):
        """
        Collect every leaf sharing the same parent
        """
        cl = []
        for e in self.__nodes:
            if e["parent"] == parent:
                cl.append(e)

        return cl

    def collectUnlockedLeaves(self, parent):
        """
        Collect every leaf sharing the same parent
        unless it is locked
        """
        cl = []
        for e in self.__nodes:
            if e["parent"] == parent:
                if not e["lock"]:
                    cl.append(e)

        return cl

    def lockNode(self, node):
        node["lock"] = 1

    #
    # Start: styles management
    #
    def styleAppend(self, stylename, key, val):
        if stylename not in self.__styles:
            self.__styles[stylename] = []

        self.__styles[stylename].append([key, val])

    def styleApply(self, stylename, node_or_link):
        node_or_link["style"] = stylename

    def styleDefaultAppend(self, key, val):
        self.__default_style.append([key, val])

    def styleDefaultLinksAppend(self, key, val):
        self.__default_style_links.append([key, val])

    #
    # End: styles management
    #

    #
    # Start: properties management
    #
    def propertiesAsStringGet(self, node, props):
        """
        Get the properties string according to parent/children
        props is the properties dictionnary
        """
        allProps = {}

        #
        # Default style come first, they can then be overriden
        #
        if self.__default_style:
            allProps.update(self.__default_style)

        #
        # First, we build the styles
        #
        if node["style"]:
            stylename = node["style"]
            allProps.update(self.__styles[stylename])

        #
        # Now we build the properties:
        # remember they override styles
        #
        allProps.update(props)

        if self.__has_children(node):
            propStringList = ["%s=%s;\n" % (k, format_property(k, v)) for k, v in allProps.items()]
            properties = "".join(propStringList)
        else:
            if props:
                propStringList = ["%s=%s" % (k, format_property(k, v)) for k, v in allProps.items()]
                properties = "[" + ",".join(propStringList) + "]"
            else:
                properties = ""

        return properties

    def propertiesLinkAsStringGet(self, link):
        props = {}

        if self.__default_style_links:
            props.update(self.__default_style_links)

        if link["style"]:
            stylename = link["style"]

            # Build the properties string for node
            props.update(self.__styles[stylename])

        props.update(link["properties"])

        properties = ""
        if props:
            properties += ",".join(["%s=%s" % (str(k), format_property(k, val)) for k, val in props.items()])
        return properties

    def propertyForeachLinksAppend(self, node, key, val):
        for l in self.__links:
            if l["from_node"] == node:
                props = l["properties"]
                props[key] = val

    def propertyAppend(self, node_or_link, key, val):
        """
        Append a property to the wanted node or link
        mynode = newItem(\"blah\")
        Ex. propertyAppend(mynode, \"color\", \"red\")
        """
        props = node_or_link["properties"]
        props[key] = val

    def propertyGet(self, node_or_link, key):
        """
        Get the value of a given property
        Ex. prop = propertyGet(node, \"color\")
        """
        try:
            props = node_or_link["properties"]
            return props[key]
        except:
            return None

    def propertyRemove(self, node_or_link, key):
        """
        Remove a property to the wanted node or link
        mynode = newItem(\"blah\")
        Ex. propertyRemove(mynode, \"color\")
        """
        props = node_or_link["properties"]
        del props[key]

    #
    # End: Properties management
    #

    #
    # For a good legend, the graph must have
    # rankdir=LR property set.
    #
    def legendAppend(self, legendstyle, legenddescr, labelin=None):
        if labelin:
            item = self.newItem(legenddescr, self.legend)
            self.styleApply(legendstyle, item)
        else:
            style = self.newItem("", self.legend)
            descr = self.newItem(legenddescr, self.legend)
            self.styleApply(legendstyle, style)
            link = self.newLink(style, descr)
            self.propertyAppend(link, "dir", "none")
            self.propertyAppend(link, "style", "invis")
            self.propertyAppend(descr, "shape", "plaintext")

    def tree_debug(self, level, node, children):
        if children:
            print("(level:%d) Eid:%d has children (%s)" % (level, node["id"], str(children)))
        else:
            print("Eid:" + str(node["id"]) + " has no children")

    #
    # Core function that outputs the data structure tree into dot language
    #
    def tree(self, level, node, children):
        """
        Core function to output dot which sorts out parents and children
        and do it in the right order
        """
        #         print('%stree(level %s, ID %s, %s)' % ('  ' * level, level, node['id'],
        #                                                 len(children) if children else 'no children'))
        if debug:
            print("/* Grabed node = %s*/" % str(node["id"]))

        if node["lock"] == 1:  # The node is locked, nothing should be printed
            if debug:
                print("/* The node (%s) is locked */" % str(node["id"]))

            if self.__opened_braces:
                self.fd.write(level * self.padding_str)
                self.fd.write("}\n")
                self.__opened_braces.pop()
            return

        props = node["properties"]

        if children:
            node["lock"] = 1
            self.fd.write(level * self.padding_str)

            # print('level %d Starting subcluster for %d %r' % (level, node['id'], node.get('label', '(no label)')))

            self.fd.write(self.padding_str + "subgraph cluster%d {\n" % node["id"])
            properties = self.propertiesAsStringGet(node, props)
            self.fd.write(level * self.padding_str)
            self.fd.write(self.padding_str + "%s" % properties)
            self.__opened_braces.append([node, level])
        else:
            # We grab appropriate properties
            properties = self.propertiesAsStringGet(node, props)
            # "node%d %s;\n" % (node['id'], properties))
            # We get the latest opened elements
            if self.__opened_braces:
                last_cluster, last_level = self.__opened_braces[-1]
            else:
                last_cluster = None
                last_level = 0

            if debug:
                if node["parent"]:
                    parent_str = str(node["parent"]["id"])
                else:
                    parent_str = "None"
                if last_cluster:
                    last_cluster_str = str(last_cluster["id"])
                else:
                    last_cluster_str = "None"
                print(
                    "/* e[parent] = %s, last_cluster = %s, last_level = %d, opened_braces: %s */"
                    % (parent_str, last_cluster_str, last_level, str(self.__opened_braces),)
                )

            # Write children/parent with properties
            if node["parent"]:
                if node["parent"] != last_cluster:
                    while node["parent"] < last_cluster:
                        last_cluster, last_level = self.__opened_braces[-1]
                        if node["parent"] == last_cluster:
                            last_level += 1
                            # We browse any property to build a string
                            self.fd.write(last_level * self.padding_str)
                            self.fd.write(self.padding_str + "node%d %s;\n" % (node["id"], properties))
                            node["lock"] = 1
                        else:
                            self.fd.write(last_level * self.padding_str)
                            self.fd.write(self.padding_str + "}\n")
                            self.__opened_braces.pop()
                else:
                    self.fd.write(level * self.padding_str)
                    self.fd.write(self.padding_str + "node%d %s;\n" % (node["id"], properties))
                    node["lock"] = 1
                    cl = self.collectUnlockedLeaves(node["parent"])
                    for l in cl:
                        props = l["properties"]
                        properties = self.propertiesAsStringGet(l, props)
                        self.fd.write(last_level * self.padding_str)
                        self.fd.write(
                            self.padding_str + self.padding_str + "node%d %s;\n" % (l["id"], properties)
                        )
                        node["lock"] = 1
                        self.lockNode(l)

                    self.fd.write(level * self.padding_str + "}\n")
                    self.__opened_braces.pop()

            else:
                self.fd.write(self.padding_str + "node%d %s;\n" % (node["id"], properties))
                node["lock"] = 1

    def browse(self, node, cb):
        """
        Browse nodes in a tree and calls cb providing node parameters
        """
        children = self.__has_children(node)
        if children:
            cb(self.__browse_level, node, str(children))
            for c in children:
                self.__browse_level += 1
                self.browse(c, cb)

        else:
            cb(self.__browse_level, node, None)
            self.__browse_level = 0

    def browse2_(self, G, id_node, cb, level):
        """
            cb(level, node_data, children_results)
        """
        children_result = {}
        children = G.successors(id_node)
        for c in children:
            rec = self.browse2_(G=G, id_node=c, cb=cb, level=level + 1)
            children_result[c] = rec
        node_data = G.nodes[id_node]["data"]
        return cb(level, node_data, children_result)

    def browse2(self, cb):
        import networkx as nx

        G = nx.DiGraph()
        G.add_node("root", data=dict(id="root"))
        for n in self.__nodes:
            G.add_node(n["id"], data=n)

        for n in self.__nodes:
            parent = n["parent"]
            if parent is None:
                id_parent = "root"
            else:
                id_parent = parent["id"]
            G.add_edge(id_parent, n["id"])

        return self.browse2_(G=G, id_node="root", cb=cb, level=0)

    def structure(self):
        def example(level, node, children_results):  # @UnusedVariable
            if children_results:
                c = ", ".join(["%s" % v for k, v in children_results.items()])  # @UnusedVariable
                return "%s { %s } " % (node["id"], c)
            else:
                return node["id"]

        return self.browse2(example)

    def dot2(self):
        # from contracts.utils import indent
        def indented_results(children_results):
            s = ""
            for cs in children_results.values():
                s += indentu(cs, "   ") + "\n"
            return s

        def render_dot_root(level, node, children_results):  # @UnusedVariable
            s = "digraph G { \n"
            s += """
                graph [fontname = "FONT"];
                 node [fontname = "FONT"];
                 edge [fontname = "FONT"];
            """.replace(
                "FONT", "Anka/Coder Narrow"
            )

            if self.options:
                for key, value in self.options.items():
                    s += "    %s=%s;" % (key, value)
                s += "\n"
            assert isinstance(s, str), s.__repr__()
            r = indented_results(children_results)  # .decode("unicode_escape")
            try:
                s += r
            except UnicodeDecodeError:
                # print r[210:220].__repr__()
                #                 print type(r), type(s)
                #                 print r[1200:1230].__repr__()
                #                 print r.__repr__()
                raise

            self.fd = StringIO()
            for n in self.__nodes:
                self.dotLinks(n)

            s += indentu(self.fd.getvalue(), "   ")
            s += "}"
            return s

        def render_dot_subgraph(level, node, children_results):  # @UnusedVariable
            s = "subgraph cluster%d { \n" % node["id"]
            properties = self.propertiesAsStringGet(node, node["properties"])
            s += indentu("   ", properties)
            s += indented_results(children_results)
            s += "}"
            return s

        def render_dot_node(level, node, children_results):  # @UnusedVariable
            properties = self.propertiesAsStringGet(node, node["properties"])
            return "node%d %s;\n" % (node["id"], properties)

        def render_dot(level, node, children_results):  # @UnusedVariable

            if node["id"] == "root":
                return render_dot_root(level, node, children_results)

            if children_results:
                return render_dot_subgraph(level, node, children_results)
            else:
                return render_dot_node(level, node, children_results)

        return self.browse2(render_dot)

    def dotLinks(self, node):
        """
        Write links between nodes
        """
        for l in self.__links:
            if l["from_node"] == node:
                # Check if we link form a cluster
                children = self.__has_children(node)
                if children:
                    if l["cl_from_node"]:
                        src = l["cl_from_node"]["id"]
                    else:
                        src = children[0]["id"]
                    cluster_src = node["id"]
                else:
                    src = node["id"]
                    cluster_src = ""

                # Check if we link to a cluster
                children = self.__has_children(l["to_node"])
                if children:
                    if l["cl_to_node"]:
                        dst = l["cl_to_node"]["id"]
                    else:
                        dst = children[0]["id"]
                    cluster_dst = l["to_node"]["id"]
                else:
                    dst = l["to_node"]["id"]
                    cluster_dst = ""

                self.fd.write("node%d->node%d" % (src, dst))

                props = self.propertiesLinkAsStringGet(l)

                # Build new properties if we link from or to a cluster
                if cluster_src:
                    if props:
                        props += ","
                    props += "ltail=cluster%d" % cluster_src
                if cluster_dst:
                    if props:
                        props += ","
                    props += "lhead=cluster%d" % cluster_dst

                if props:
                    self.fd.write(" [%s]" % props)

                self.fd.write(";\n")

    def dot(self, fd=stdout):
        """
        Translates the datastructure into dot
        """
        try:
            self.fd = fd

            #             self.fd.write("/* Generated by GvGen v.%s (http://www.picviz.com/sections/opensource/gvgen.html) */\n\n" % (gvgen_version))

            self.fd.write("digraph G {\n")

            if self.options:
                for key, value in self.options.items():
                    self.fd.write("%s=%s;" % (key, value))
                self.fd.write("\n")

            # We write parents and children in order
            for e in self.__nodes:
                if debug_tree_unroll:
                    self.browse(e, self.tree_debug)
                else:
                    self.browse(e, self.tree)

            # We write the connection between nodes
            for e in self.__nodes:
                self.dotLinks(e)

            # We put all the nodes belonging to the parent
            self.fd.write("}\n")
        finally:
            # Remove our reference to file descriptor
            self.fd = None

    # XXX: unused
    # def remove_identity_nodes(self):
    #     """
    #         Certainly need to remove ADD_ORDER = True

    #     """
    #     def is_identity(n):
    #         return n['properties']['label'] == 'Identity'

    #     import networkx as nx
    #     G = nx.DiGraph()

    #     leqs = set()
    #     identities = set()
    #     for n in self.__nodes:
    #         if is_identity(n):
    #             identities.add(n['id'])
    #         if n['style'] == 'leq':
    #             leqs.add(n['id'])

    #     for l in self.__links:
    #         from_node = l['from_node']['id']
    #         to_node = l['to_node']['id']
    #         G.add_edge(from_node, to_node)

    #     def remove_link(tail , head):
    #         def matches(l):
    #             return (l['to_node']['id'] == head) and(l['from_node']['id'] == tail)
    #         links = []
    #         nmatches = 0
    #         for l in self.__links:
    #             if not matches(l):
    #                 links.append(l)
    #                 print l['from_node']['id'], l['to_node']['id']
    #             else:
    #                 nmatches += 1
    #         if not nmatches:
    #             raise ValueError('No link %s - %s found' % (tail, head))
    #         self.__links = links

    #     def remove_node(n):
    #         def matches(x):
    #             return x['id'] == n
    #         self.__nodes = [x for x in self.__nodes if not matches(x)]

    #         for l in self.__links:
    #             if l['to_node']['id'] == n:
    #                 raise ValueError('Found link to node %r' % n)
    #             if l['from_node']['id'] == n:
    #                 raise ValueError('Found link from node %r' % n)

    #     def get_node(n):
    #         for x in self.__nodes:
    #             if x['id'] == n:
    #                 return x

    #     def sub_link_from(old, new):
    #         for l in self.__links:
    #             if l['from_node']['id'] == old:
    #                 l['from_node'] = get_node(new)

    #     # look for A -l1--> [prev_leq] -l2-> [Identity n] --l3-> [B]
    #     # look for A -l3--> [B]
    #     for n in identities:
    #         pred = G.predecessors(n)
    #         has_prev_leq = len(pred) == 1 and pred[0] in leqs
    #         succ = G.successors(n)
    #         has_one_succ = len(succ) == 1
    #         has_prev_leq_one = has_prev_leq and len(G.predecessors(pred[0])) == 1

    #         if not has_prev_leq_one or not has_one_succ:
    #             print('This one is not eligible (prev_eq')
    #             continue

    #         prev_leq = pred[0]
    #         A = G.predecessors(prev_leq)[0]

    #         # we also want the previous ones to be

    #         remove_link(A, prev_leq)  # l1
    #         remove_link(prev_leq, n)  # l2
    #         sub_link_from(old=n, new=A)

    #         print 'n', n, 'pred', pred, 'succ', succ
    #         remove_node(n)
    #         remove_node(prev_leq)


#         self.__links = []


def format_property(k, v):  # @UnusedVariable
    s = str(v)

    def escape(x):
        return x.replace("\n", "\\n")

    if s and s[0] == "<":
        res = "<%s>" % escape(s)
    else:
        res = '"%s"' % escape(s)
    return res


if __name__ == "__main__":
    graph = ACGvGen()

    graph.smart_mode = 1

    graph.styleDefaultAppend("color", "blue")

    parents = graph.newItem("Parents")
    father = graph.newItem("Bob", parents)
    mother = graph.newItem("Alice", parents)
    children = graph.newItem("Children")
    child1 = graph.newItem("Carol", children)
    child2 = graph.newItem("Eve", children)
    child3 = graph.newItem("Isaac", children)
    postman = graph.newItem("Postman")
    graph.newLink(father, child1)
    graph.newLink(child1, father)
    graph.newLink(father, child1)
    graph.newLink(father, child2)
    graph.newLink(mother, child2)
    myl = graph.newLink(mother, child1)
    graph.newLink(mother, child3)
    graph.newLink(postman, child3, "Email is safer")
    graph.newLink(parents, postman)  # Cluster link

    graph.propertyForeachLinksAppend(parents, "color", "blue")

    graph.propertyForeachLinksAppend(father, "label", "My big link")
    graph.propertyForeachLinksAppend(father, "color", "red")

    graph.propertyAppend(postman, "color", "red")
    graph.propertyAppend(postman, "fontcolor", "white")

    graph.styleAppend("link", "label", "mylink")
    graph.styleAppend("link", "color", "green")
    graph.styleApply("link", myl)
    graph.propertyAppend(myl, "arrowhead", "empty")

    graph.styleAppend("Post", "color", "blue")
    graph.styleAppend("Post", "style", "filled")
    graph.styleAppend("Post", "shape", "rectangle")
    graph.styleApply("Post", postman)

    graph.dot()


def indentu(s, prefix, first=None):
    assert isinstance(prefix, str)
    lines = s.split("\n")
    if not lines:
        return ""

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = " " * (m - len(prefix)) + prefix
    first = " " * (m - len(first)) + first

    # differnet first prefix
    res = ["%s%s" % (prefix, line.rstrip()) for line in lines]
    res[0] = "%s%s" % (first, lines[0].rstrip())
    return "\n".join(res)



# coding=utf-8
from .dynamics import *
from .lane_pose import *
from .measurements import *
from .segment import *
from .svg import *
from .tags import *
from .world_building import *
from .sampling_poses import *
from .pwm_dynamics import *
from .lane_at_point import *
from .typing_tests import *

#
# def jobs_comptests(context):
#     # instantiation
#     # from comptests import jobs_registrar
#     from comptests.registrar import jobs_registrar_simple
#     jobs_registrar_simple(context)



# coding=utf-8
import math

import numpy as np
from duckietown_serialization_ds1 import Serializable

from .region import Region

__all__ = ["RectangularArea"]


class RectangularArea(Serializable, Region):
    def __init__(self, pmin, pmax):
        self.pmin = np.array(pmin, dtype=np.float64)
        self.pmax = np.array(pmax, dtype=np.float64)

        if not np.all(self.pmin < self.pmax):
            msg = "Invalid area: %s %s" % (pmin, pmax)
            raise ValueError(msg)

    @classmethod
    def join(cls, a, b):
        pmin = np.minimum(a.pmin, b.pmin)
        pmax = np.maximum(a.pmax, b.pmax)
        return RectangularArea(pmin, pmax)

    def contains(self, p, epsilon=0.0):
        pmin, pmax = self.pmin, self.pmax
        return (pmin[0] - epsilon <= p[0] <= pmax[0] + epsilon) and (
            pmin[1] - epsilon <= p[1] <= pmax[1] + epsilon
        )

    def distance(self, p) -> float:
        pmin, pmax = self.pmin, self.pmax
        abs = math.fabs
        if pmin[0] <= p[0] <= pmax[0]:
            d0 = 0.0
        else:
            d0 = min(abs(pmin[0] - p[0]), abs(pmax[0] - p[0]))

        if pmin[1] <= p[1] <= pmax[1]:
            d1 = 0.0
        else:
            d1 = d0 = min(abs(pmin[1] - p[1]), abs(pmax[1] - p[1]))

        return math.hypot(d0, d1)



# coding=utf-8
import numpy as np
from contracts import contract

import geometry as geo

__all__ = ["SE2_interpolate", "SE2_apply_R2"]


@contract(q0="SE2", q1="SE2", alpha="int|(float,finite)")
def SE2_interpolate(q0, q1, alpha):
    alpha = float(alpha)
    v = geo.SE2.algebra_from_group(geo.SE2.multiply(geo.SE2.inverse(q0), q1))
    vi = v * alpha
    q = np.dot(q0, geo.SE2.group_from_algebra(vi))
    return q


@contract(q="SE2", p="R2")
def SE2_apply_R2(q, p):
    pp = [p[0], p[1], 1]
    res = np.dot(q, pp)
    return res[:2]



# coding=utf-8
import itertools
import os
import traceback
from typing import List

import numpy as np
import oyaml as yaml

import geometry as geo
from duckietown_serialization_ds1 import Serializable
from .duckiebot import DB18
from .duckietown_map import DuckietownMap
from .other_objects import (
    Barrier,
    Building,
    Bus,
    Cone,
    Duckie,
    GenericObject,
    House,
    SIGNS,
    Tree,
    Truck,
)
from .tags_db import FloorTag
from .tile import Tile
from .tile_map import TileMap
from .tile_template import load_tile_types
from .traffic_light import TrafficLight
from .. import logger
from ..geo import Scale2D, SE2Transform
from ..geo.measurements_utils import iterate_by_class

__all__ = ["create_map", "list_maps", "construct_map", "load_map"]


def create_map(H=3, W=3) -> TileMap:
    tile_map = TileMap(H=H, W=W)

    for i, j in itertools.product(range(H), range(W)):
        tile = Tile(kind="mykind", drivable=True)
        tile_map.add_tile(i, j, "N", tile)

    return tile_map


def list_maps() -> List[str]:
    maps_dir = get_maps_dir()

    def f():
        for map_file in os.listdir(maps_dir):
            map_name = map_file.split(".")[0]
            yield map_name

    return list(f())


def get_maps_dir() -> str:
    abs_path_module = os.path.realpath(__file__)
    module_dir = os.path.dirname(abs_path_module)
    d = os.path.join(module_dir, "../data/gd1/maps")
    assert os.path.exists(d), d
    return d


def get_texture_dirs() -> List[str]:
    abs_path_module = os.path.realpath(__file__)
    module_dir = os.path.dirname(abs_path_module)
    d = os.path.join(module_dir, "../data/gd1/textures")
    assert os.path.exists(d), d
    d2 = os.path.join(module_dir, "../data/gd1/meshes")
    assert os.path.exists(d2), d2

    d3 = os.path.join(module_dir, "../data/tag36h11")
    assert os.path.exists(d3), d3

    return [d, d2, d3]


def _get_map_yaml(map_name: str) -> str:
    maps_dir = get_maps_dir()
    fn = os.path.join(maps_dir, map_name + ".yaml")
    if not os.path.exists(fn):
        msg = "Could not find file %s" % fn
        raise ValueError(msg)
    with open(fn) as _:
        data = _.read()
    return data


def load_map(map_name: str) -> DuckietownMap:
    logger.info("loading map %s" % map_name)
    data = _get_map_yaml(map_name)
    yaml_data = yaml.load(data, Loader=yaml.SafeLoader)

    return construct_map(yaml_data)


def construct_map(yaml_data: dict) -> DuckietownMap:
    tile_size = yaml_data["tile_size"]
    dm = DuckietownMap(tile_size)
    tiles = yaml_data["tiles"]
    assert len(tiles) > 0
    assert len(tiles[0]) > 0

    # Create the grid
    A = len(tiles)
    B = len(tiles[0])
    tm = TileMap(H=B, W=A)

    templates = load_tile_types()
    for a, row in enumerate(tiles):
        if len(row) != B:
            msg = "each row of tiles must have the same length"
            raise ValueError(msg)

        # For each tile in this row
        for b, tile in enumerate(row):
            tile = tile.strip()

            if tile == "empty":
                continue

            DEFAULT_ORIENT = "E"  # = no rotation
            if "/" in tile:
                kind, orient = tile.split("/")
                kind = kind.strip()
                orient = orient.strip()

                drivable = True
            elif "4" in tile:
                kind = "4way"
                # angle = 2
                orient = DEFAULT_ORIENT
                drivable = True
            else:
                kind = tile
                # angle = 0
                orient = DEFAULT_ORIENT
                drivable = False

            tile = Tile(kind=kind, drivable=drivable)
            if kind in templates:
                tile.set_object(kind, templates[kind], ground_truth=SE2Transform.identity())
            else:
                pass
                # msg = 'Could not find %r in %s' % (kind, templates)
                # logger.debug(msg)

            tm.add_tile(b, (A - 1) - a, orient, tile)

    def go(obj_name, desc):
        obj = get_object(desc)
        transform = get_transform(desc, tm, tile_size)
        dm.set_object(obj_name, obj, ground_truth=transform)

    objects = yaml_data.get("objects", [])
    if isinstance(objects, list):
        for obj_idx, desc in enumerate(objects):
            kind = desc["kind"]
            obj_name = "ob%02d-%s" % (obj_idx, kind)
            go(obj_name, desc)
    elif isinstance(objects, dict):
        for obj_name, desc in objects.items():
            go(obj_name, desc)
    else:
        raise ValueError(objects)

    for it in list(iterate_by_class(tm, Tile)):
        ob = it.object
        if "slots" in ob.children:
            slots = ob.children["slots"]
            for k, v in list(slots.children.items()):
                if not v.children:
                    slots.remove_object(k)
            if not slots.children:
                ob.remove_object("slots")

    dm.set_object("tilemap", tm, ground_truth=Scale2D(tile_size))
    return dm


def get_object(desc):
    kind = desc["kind"]

    attrs = {}
    if "tag" in desc:
        tag_desc = desc["tag"]
        # tag_id = tag_desc.get('tag_id')
        # size = tag_desc.get('size', DEFAULT_TAG_SIZE)
        # family = tag_desc.get('family', DEFAULT_FAMILY)
        attrs["tag"] = Serializable.from_json_dict(tag_desc)

    kind2klass = {
        "trafficlight": TrafficLight,
        "duckie": Duckie,
        "cone": Cone,
        "barrier": Barrier,
        "building": Building,
        "duckiebot": DB18,
        "tree": Tree,
        "house": House,
        "bus": Bus,
        "truck": Truck,
        "floor_tag": FloorTag,
    }
    kind2klass.update(SIGNS)
    if kind in kind2klass:
        klass = kind2klass[kind]
        try:
            obj = klass(**attrs)
        except TypeError:
            msg = "Could not initialize %s with attrs %s:\n%s" % (
                klass.__name__,
                attrs,
                traceback.format_exc(),
            )
            raise Exception(msg)

    else:
        logger.debug("Do not know special kind %s" % kind)
        obj = GenericObject(kind=kind)
    return obj


def get_transform(desc, tm, tile_size):
    rotate_deg = desc.get("rotate", 0)
    rotate = np.deg2rad(rotate_deg)

    if "pos" in desc:
        pos = desc["pos"]
        x = float(pos[0]) * tile_size
        # account for non-righthanded
        y = float(tm.W - pos[1]) * tile_size
        # account for non-righthanded
        rotate = -rotate
        transform = SE2Transform([x, y], rotate)
        return transform

    else:

        if "pose" in desc:
            pose = Serializable.from_json_dict(desc["pose"])
        else:
            pose = SE2Transform.identity()

        if "attach" in desc:
            attach = desc["attach"]
            tile_coords = tuple(attach["tile"])
            slot = str(attach["slot"])

            x, y = get_xy_slot(slot)
            i, j = tile_coords

            u, v = (x + i) * tile_size, (y + j) * tile_size
            transform = SE2Transform([u, v], rotate)

            q = geo.SE2.multiply(transform.as_SE2(), pose.as_SE2())

            return SE2Transform.from_SE2(q)
        else:
            return pose


def get_xy_slot(i):
    # tile_offset
    # to = 0.20
    # # tile_curb
    # tc = 0.05
    to = 0.09  # [m] the longest of two distances from the center of the tag to the edge
    tc = 0.035  # [m] the shortest of two distances from the center of the tag to the edge

    positions = {
        0: (+to, +tc),
        1: (+tc, +to),
        2: (-tc, +to),
        3: (-to, +tc),
        4: (-to, -tc),
        5: (-tc, -to),
        6: (+tc, -to),
        7: (+to, -tc),
    }
    x, y = positions[int(i)]
    return x, y


def get_texture_file(tex_name: str) -> str:
    res = []
    tried = []
    for d in get_texture_dirs():
        suffixes = ["", "_1", "_2", "_3", "_4"]
        for s in suffixes:
            for ext in [".jpg", ".png", ""]:
                path = os.path.join(d, tex_name + s + ext)
                tried.append(path)
                if os.path.exists(path):
                    res.append(path)

    if not res:
        msg = "Could not find any texture for %s" % tex_name
        logger.debug("tried %s" % tried)
        raise KeyError(msg)
    return res[0]



from collections import defaultdict
from dataclasses import dataclass
from typing import cast, Dict, Optional, Set, Tuple

import numpy as np
from networkx import DiGraph, MultiDiGraph

import geometry as geo
from duckietown_world.geo import Matrix2D, PlacedObject, SE2Transform
from duckietown_world.geo.measurements_utils import iterate_by_class
from zuper_commons.types import ZException
from .duckietown_map import DuckietownMap
from .lane_segment import LaneSegment
from .tile_coords import TileCoords

import networkx as nx


@dataclass
class SkeletonGraphResult:
    root: PlacedObject
    root2: PlacedObject
    G: MultiDiGraph
    # This is a graph with nodes PointLabels
    G0: DiGraph


__all__ = ["get_skeleton_graph", "SkeletonGraphResult"]

PointLabel = Tuple[float, float, float, float]


@dataclass
class MeetingPoint:
    incoming: Set[str]  # set of lane names
    outcoming: Set[str]  # set of lanes

    connects_to: Set[PointLabel]  # set of points
    point: Optional[SE2Transform]
    into_tile: Optional[Tuple[int, int]]
    from_tile: Optional[Tuple[int, int]]
    #
    # def __init__(self):
    #     self.point = None
    #     self.incoming = set()
    #     self.outcoming = set()
    #     self.connects_to = set()
    #     self.into_tile = None
    #     self.from_tile = None

    # def __repr__(self):
    #     return "MP(%d %d | %s, %s)" % (
    #         len(self.incoming),
    #         len(self.outcoming),
    #         self.incoming,
    #         self.outcoming,
    #     )


def discretize(tran: SE2Transform) -> PointLabel:
    def D(x):
        return np.round(x, decimals=2)

    p, theta = geo.translation_angle_from_SE2(tran.as_SE2())
    return D(p[0]), D(p[1]), D(np.cos(theta)), D(np.sin(theta))


def graph_for_meeting_points(mp: Dict[str, MeetingPoint]) -> DiGraph:
    G = DiGraph()
    for k, p in mp.items():
        G.add_node(k, meeting_point=p)
    for k, p in mp.items():
        for k2 in p.connects_to:
            G.add_edge(k, k2)
    return G


def get_skeleton_graph(po: DuckietownMap) -> SkeletonGraphResult:
    """ Returns a graph with the lane segments of the map """

    root = PlacedObject()

    meeting_points: Dict[str, MeetingPoint] = defaultdict(MeetingPoint)

    for i, it in enumerate(iterate_by_class(po, LaneSegment)):
        lane_segment = cast(LaneSegment, it.object)
        assert isinstance(lane_segment, LaneSegment), lane_segment

        absolute_pose = it.transform_sequence.asmatrix2d()

        lane_segment_transformed = transform_lane_segment(lane_segment, absolute_pose)

        identity = SE2Transform.identity()
        name = "ls%03d" % i
        root.set_object(name, lane_segment_transformed, ground_truth=identity)

        p0 = discretize(lane_segment_transformed.control_points[0])
        p1 = discretize(lane_segment_transformed.control_points[-1])

        if not p0 in meeting_points:
            meeting_points[p0] = MeetingPoint(
                set(), set(), set(), lane_segment_transformed.control_points[0], None, None,
            )
        if not p1 in meeting_points:
            meeting_points[p1] = MeetingPoint(
                set(), set(), set(), lane_segment_transformed.control_points[-1], None, None,
            )

        # meeting_points[p0].point = lane_segment_transformed.control_points[0]
        meeting_points[p0].outcoming.add(name)
        # meeting_points[p1].point = lane_segment_transformed.control_points[-1]
        meeting_points[p1].incoming.add(name)

        meeting_points[p0].connects_to.add(p1)

        tile_coords = [_ for _ in it.transform_sequence.transforms if isinstance(_, TileCoords)]
        if not tile_coords:
            raise ZException(p0=p0, p1=p1, transforms=it.transform_sequence.transforms)
        tile_coord = tile_coords[0]
        ij = tile_coord.i, tile_coord.j
        meeting_points[p0].into_tile = ij
        meeting_points[p1].from_tile = ij

    for k, mp in meeting_points.items():
        if (len(mp.incoming) == 0) or (len(mp.outcoming) == 0):
            msg = "Completeness assumption violated at point %s: %s" % (k, mp)
            raise Exception(msg)

    G0 = graph_for_meeting_points(meeting_points)
    # compress the lanes which are contiguous

    aliases = {}

    created = {}

    def resolve_alias(x):
        return x if x not in aliases else resolve_alias(aliases[x])

    for k, mp in list(meeting_points.items()):
        # continue
        if not (len(mp.incoming) == 1 and len(mp.outcoming) == 1):
            continue

        # not necessary anymore
        meeting_points.pop(k)
        lin_name = list(mp.incoming)[0]
        lout_name = list(mp.outcoming)[0]

        lin_name = resolve_alias(lin_name)
        lout_name = resolve_alias(lout_name)

        # print(' -> %s and %s meet at %s' % (lin_name, lout_name, mp))
        # print('%s and %s meet at %s' % (lin_name, lout_name, k))

        def get(it):
            if it in root.children:
                return root.children[it]
            else:
                return created[it]

        lin = get(lin_name)
        lout = get(lout_name)

        # name = 'alias%s' % (len(aliases))
        # name = '%s-%s' % (lin_name, lout_name)
        name = "L%d" % (len(created))
        width = lin.width

        control_points = lin.control_points + lout.control_points[1:]
        ls = LaneSegment(width=width, control_points=control_points)
        created[name] = ls

        aliases[lin_name] = name
        aliases[lout_name] = name
        # print('new alias %s' % name)
    #
    # print('created: %s' % list(created))
    # print('aliases: %s' % aliases)
    root2 = PlacedObject()
    for k, v in created.items():
        if not k in aliases:
            root2.set_object(k, v, ground_truth=SE2Transform.identity())

    for k, v in root.children.items():
        if not k in aliases:
            root2.set_object(k, v, ground_truth=SE2Transform.identity())

    G = nx.MultiDiGraph()

    k2name = {}
    for i, (k, mp) in enumerate(meeting_points.items()):
        node_name = "P%d" % i
        k2name[k] = node_name

        G.add_node(node_name, point=mp.point)

    ls2start = {}
    ls2end = {}
    for i, (k, mp) in enumerate(meeting_points.items()):
        node_name = k2name[k]

        for l in mp.incoming:
            ls2end[resolve_alias(l)] = node_name
        for l in mp.outcoming:
            ls2start[resolve_alias(l)] = node_name

    # print(ls2start)
    # print(ls2end)

    for l in ls2start:
        n1 = ls2start[l]
        n2 = ls2end[l]
        G.add_edge(n1, n2, lane=l)

    return SkeletonGraphResult(root=root, root2=root2, G=G, G0=G0)


def transform_lane_segment(lane_segment: LaneSegment, transformation: Matrix2D) -> LaneSegment:
    M = transformation.m

    def transform_point(p):
        q = p.as_SE2()
        q2 = np.dot(M, q)
        p2 = SE2Transform.from_SE2(q2)
        return p2

    control_points = list(map(transform_point, lane_segment.control_points))

    det = np.linalg.det(M)
    width = float(lane_segment.width * np.sqrt(det))
    return LaneSegment(control_points=control_points, width=width)



# coding=utf-8
from duckietown_serialization_ds1 import Serializable

import geometry as geo
from .platform_dynamics import PlatformDynamicsFactory, PlatformDynamics
from .types import *

__all__ = ["GenericKinematicsSE2"]

# noinspection PyUnresolvedReferences
from geometry.poses import *


class GenericKinematicsSE2(PlatformDynamicsFactory, PlatformDynamics, Serializable):
    """
        Any dynamics on SE(2)

        Commands = velocities in se(2)
    """

    @classmethod
    def initialize(cls, c0: TSE2v, t0: float = 0, seed=None):
        return GenericKinematicsSE2(c0, t0)

    def __init__(self, c0: TSE2v, t0: float):
        # start at q0, v0
        q0, v0 = c0
        geo.SE2.belongs(q0)
        geo.se2.belongs(v0)
        self.t0 = t0
        self.v0 = v0
        self.q0 = q0

    def integrate(self, dt: float, commands: se2v) -> "GenericKinematicsSE2":
        """ commands = velocity in body frame """
        # convert to float
        dt = float(dt)
        # the commands must belong to se(2)
        geo.se2.belongs(commands)
        v = commands
        # suppose we hold v for dt, which pose are we going to?
        diff = geo.SE2.group_from_algebra(dt * v)  # exponential map
        # compute the absolute new pose; applying diff from q0
        q1 = geo.SE2.multiply(self.q0, diff)
        # the new configuration
        c1 = q1, v
        # the new time
        t1 = self.t0 + dt
        # return the new state
        return GenericKinematicsSE2(c1, t1)

    def TSE2_from_state(self) -> TSE2v:
        return self.q0, self.v0



import plotly.offline as offline

res = offline.plot(
    {"data": [{"y": [4, 2, 3, 4]}], "layout": {"title": "Test Plot", "font": dict(size=16)},},
    output_type="div",
)

with open("testplotly.html", "w") as f:
    f.write('<div style="width: 600px; height: 300">')
    f.write(res)
    f.write("</div>")



# coding=utf-8
import sys

from setuptools import find_packages, setup


def get_version(filename):
    import ast

    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename="src/duckietown_world/__init__.py")

install_requires = [
    "numpy",
    "PyYAML",
    "networkx>=2.2,<3",
    "duckietown-serialization-ds1<2",
    "svgwrite",
    "PyGeometry-z6",
    "beautifulsoup4>=4.6.3",
    "lxml",
    "pillow",
    "future",
    "scipy",
    "plotly",
    "oyaml",
    "markdown",
    "zuper-typing-z6>=6.0.66",
]
tests_require = ["comptests-z6"]
system_version = tuple(sys.version_info)[:3]

if system_version < (3, 7):
    install_requires.append("dataclasses")

line = "daffy"

setup(
    name=f"duckietown-world-{line}",
    version=version,
    download_url="http://github.com/duckietown/duckietown-world/tarball/%s" % version,
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=install_requires,
    tests_require=tests_require,
    # This avoids creating the egg file, which is a zip file, which makes our data
    # inaccessible by dir_from_package_name()
    zip_safe=False,
    # without this, the stuff is included but not installed
    include_package_data=True,
    entry_points={
        "console_scripts": [
            # 'dt-world-draw-log = duckietown_world.svg_drawing:draw_logs_main',
            "dt-world-draw-maps = duckietown_world.svg_drawing:draw_maps_main",
        ]
    },
)



from dataclasses import dataclass
from typing import cast, Dict, Set, Tuple

import numpy as np
from geometry import translation_from_SE2
from networkx import connected_components

from duckietown_world import get_skeleton_graph, TileMap, logger
from duckietown_world.geo.measurements_utils import (
    iterate_by_class,
    IterateByTestResult,
)
from duckietown_world.geo.placed_object import get_child_transform
from duckietown_world.geo.transforms import SE2value, Transform
from duckietown_world.world_duckietown.tile_map import ij_from_tilename
from zuper_commons.types import ZException
from zuper_typing import debug_print
from zuper_commons.types import ZValueError
from .duckietown_map import DuckietownMap
from .tile import Tile, translation_from_O3
from .utils import relative_pose
from .tile_coords import TileCoords


class NotInMap(ZException):
    pass


def get_tile_at_point(dw: DuckietownMap, q: SE2value) -> TileCoords:
    if not isinstance(q, np.ndarray):
        raise TypeError(type(q))
    l = list(iterate_by_class(dw, Tile))
    if not l:
        msg = "No tiles"
        raise ZValueError(msg, dw=dw)

    distances = {}
    for it in l:
        assert isinstance(it, IterateByTestResult), it
        assert isinstance(it.object, Tile), it.object
        tile = it.object
        tile_transform = it.transform_sequence
        for _ in tile_transform.transforms:
            if isinstance(_, TileCoords):
                tile_coords = _
                break
        else:
            msg = "Could not find tile coords in %s" % tile_transform
            assert False, msg

        # print('tile_transform: %s' % tile_transform.asmatrix2d().m)
        tile_relative_pose = relative_pose(tile_transform.asmatrix2d().m, q)
        p = translation_from_O3(tile_relative_pose)
        footprint = tile.get_footprint()
        d = footprint.distance(p)
        distances[it.fqn] = d
        # print(f'tile_relative_pose: {tile_coords} {p} {d}')
        # if d > 0.001:
        if tile.get_footprint().contains(p, 0.001):
            return tile_coords
    raise NotInMap(q=q, distances=distances)


@dataclass
class ComponentFootprint:
    # tile_names: List[str]
    tile_coords: Set[Tuple[int, int]]


def get_map_components(m: DuckietownMap) -> Dict[str, ComponentFootprint]:
    from duckietown_world.world_duckietown.segmentify import MeetingPoint

    sk2 = get_skeleton_graph(m)
    G0 = sk2.G0
    c = list(connected_components(G0.to_undirected()))
    res = {}
    for i, nodes in enumerate(c):
        s = "comp%d" % i
        tiles: Set[Tuple[int, int]] = set()
        for n in nodes:
            mp: MeetingPoint = G0.nodes[n]["meeting_point"]
            tiles.add(mp.from_tile)
            tiles.add(mp.into_tile)
        res[s] = ComponentFootprint(tiles)
    return res


#
# def compute_extent_of_tiles()
#     # compute region for tiles
#     tilemap =
#     extents = set()
#     for ij in tiles:
def contained(a: Tuple[int, int], b: Tuple[int, int], c: Tuple[int, int]) -> bool:
    return (a[0] <= b[0] <= c[0]) and (a[1] <= b[1] <= c[1])


def get_interest_map(m: DuckietownMap, q: SE2value) -> DuckietownMap:
    """ Returns the map of interest given a pose. """
    components = get_map_components(m)
    r = get_tile_at_point(m, q)
    tile = r.i, r.j
    for component_id, component in components.items():
        if tile in component.tile_coords:
            break
    else:
        msg = "Cannot find any component for this point."
        raise ZValueError(msg, tile=tile, q=q, components=components)

    tiles: Set[Tuple[int, int]] = component.tile_coords
    tmin: Tuple[int, int] = min(tiles)
    tmax: Tuple[int, int] = max(tiles)
    M = 1
    tmin = (tmin[0] - M, tmin[1] - M)
    tmax = (tmax[0] + M, tmax[1] + M)

    print(f"tmin tmax {tmin} {tmax}")
    children2 = {}
    for name, child in m.children.items():
        if name == "tilemap":
            tilemap = cast(TileMap, child)
            tiles2 = {}
            for tilename, tile in tilemap.children.items():
                coords = ij_from_tilename(tilename)
                if contained(tmin, coords, tmax):
                    tiles2[tilename] = tile

            sr2 = {}
            for sr_id, sr in list(tilemap.spatial_relations.items()):
                first = sr.b[0]
                if first in tiles2:
                    sr2[sr_id] = sr
            tilemap2 = TileMap(H=tilemap.H, W=tilemap.W, children=tiles2, spatial_relations=sr2)
            children2[name] = tilemap2
        else:
            t: Transform = get_child_transform(m, name)
            try:
                tile_for_child = get_tile_at_point(m, t.as_SE2())
            except NotInMap:
                print(f"cutting {name}")
                continue
            c = tile_for_child.i, tile_for_child.j
            if contained(tmin, c, tmax):
                print(f"keep {name} {c}")
                children2[name] = child
            else:
                print(f"cutting {name}")
    print(debug_print(children2))
    sr2 = {}
    for sr_id, sr in list(m.spatial_relations.items()):
        first = sr.b[0]
        if first in children2:
            sr2[sr_id] = sr
    return DuckietownMap(tile_size=m.tile_size, children=children2, spatial_relations=sr2)



# coding=utf-8
from typing import cast

from duckietown_world.geo import PlacedObject, SE2Transform

__all__ = ["DuckietownMap"]


class DuckietownMap(PlacedObject):
    tile_size: float

    def __init__(self, tile_size: float, *args, **kwargs):
        self.tile_size = tile_size
        PlacedObject.__init__(self, *args, **kwargs)

    def params_to_json_dict(self):
        return dict(tile_size=self.tile_size)

    def se2_from_curpos(self, cur_pos, cur_angle):
        """ Conversion from Duckietown Gym Simulator coordinates z"""
        from duckietown_world import TileMap

        tilemap: TileMap = cast(TileMap, self.children["tilemap"])
        H = tilemap.H
        gx, gy, gz = cur_pos
        p = [gx, (H - 1) * self.tile_size - gz]
        transform = SE2Transform(p, cur_angle)
        return transform

    def get_drawing_children(self):
        children = sorted(self.children)
        children.remove("tilemap")
        return ["tilemap"] + children



# coding=utf-8
from duckietown_serialization_ds1.serialization1 import as_json_dict
from duckietown_world.geo import PlacedObject
from duckietown_world.seqs import Constant
from duckietown_world.seqs.tsequence import GenericSequence

__all__ = ["TrafficLight"]
# TrafficLightStatus = str


class TrafficLight(PlacedObject):
    # status: GenericSequence[TrafficLightStatus]

    def __init__(self, status=None, **kwargs):
        # if status is None:
        #     status = Constant[TrafficLightStatus]("off")
        PlacedObject.__init__(self, **kwargs)
        self.status = status

    def params_to_json_dict(self):
        return dict(status=as_json_dict(self.status))



# coding=utf-8
import base64
import itertools
import logging
import math
import os
from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Tuple

import svgwrite
from bs4 import BeautifulSoup, Tag
from duckietown_world import logger
from duckietown_world.geo import (
    get_extent_points,
    get_static_and_dynamic,
    PlacedObject,
    RectangularArea,
)
from duckietown_world.seqs import SampledSequence, UndefinedAtTime
from duckietown_world.seqs.tsequence import Timestamp
from duckietown_world.utils import memoized_reset
from past.builtins import reduce
from PIL import Image
from six import BytesIO

__all__ = [
    "draw_recursive",
    "get_basic_upright2",
    "draw_static",
    "draw_axes",
    "draw_children",
    "data_encoded_for_src",
    "TimeseriesPlot",
]


def get_basic_upright2(filename: str, area: RectangularArea, size=(1024, 768)):
    drawing = svgwrite.Drawing(filename, size=size, debug=False)

    origin = area.pmin
    other = area.pmax
    space = other - origin

    sw = size[0] * 1.0 / space[0]
    sh = size[1] * 1.0 / space[1]

    s = min(sh, sw)

    tx = 0 - origin[0] * s
    ty = (space[1] + origin[1]) * s

    base = drawing.g(transform="translate(%s, %s) scale(%s) scale(+1,-1)" % (tx, ty, s))
    base.attribs["id"] = "base"
    i0 = int(math.floor(area.pmin[0]))
    j0 = int(math.floor(area.pmin[1]))
    i1 = int(math.ceil(area.pmax[0]))
    j1 = int(math.ceil(area.pmax[1]))

    grid = drawing.g(id="grid")
    for i, j in itertools.product(range(i0, i1), range(j0, j1)):
        where = (i, j)
        rect = drawing.rect(insert=where, fill="none", size=(1, 1), stroke_width=0.01, stroke="#eeeeee")
        grid.add(rect)

    base.add(grid)
    draw_axes(drawing, base, L=0.5, stroke_width=0.03)

    inside = drawing.g()
    inside.attribs["id"] = "inside"
    tofill = drawing.g()
    tofill.attribs["id"] = "tofill"
    inside.add(tofill)

    view = drawing.rect(
        insert=area.pmin.tolist(),
        size=(area.pmax - area.pmin).tolist(),
        stroke_width=0.01,
        fill="none",
        stroke="black",
    )
    view.attribs["id"] = "view"
    inside.add(view)
    base.add(inside)

    drawing.add(base)
    return drawing, tofill


def draw_recursive(drawing, po, g, draw_list=()):
    if () in draw_list:
        po.draw_svg(drawing, g)
    draw_children(drawing, po, g, draw_list=draw_list)


def draw_children(drawing, po, g, draw_list=()):
    for child_name in po.get_drawing_children():
        child = po.children[child_name]
        transforms = [_ for _ in po.spatial_relations.values() if _.a == () and _.b == (child_name,)]
        if transforms:

            rlist = recurive_draw_list(draw_list, child_name)

            if rlist:
                M = transforms[0].transform.asmatrix2d().m
                svg_transform = "matrix(%s,%s,%s,%s,%s,%s)" % (
                    M[0, 0],
                    M[1, 0],
                    M[0, 1],
                    M[1, 1],
                    M[0, 2],
                    M[1, 2],
                )

                g2 = drawing.g(id=child_name, transform=svg_transform)
                classes = get_typenames_for_class(child)
                if classes:
                    g2.attribs["class"] = " ".join(classes)
                draw_recursive(drawing, child, g2, draw_list=rlist)

                g.add(g2)


def get_typenames_for_class(ob):
    mro = type(ob).mro()
    names = [_.__name__ for _ in mro]
    for n in ["Serializable", "Serializable0", "PlacedObject", "object"]:
        names.remove(n)
    return names


def recurive_draw_list(draw_list, prefix):
    res = []
    for _ in draw_list:
        if _ and _[0] == prefix:
            res.append(_[1:])
    return res


def draw_static(
    root: PlacedObject,
    output_dir: str,
    pixel_size: Tuple[int, int] = (480, 480),
    area=None,
    images=None,
    timeseries=None,
    height_of_stored_images: Optional[int] = None,
    main_robot_name: Optional[str] = None,
) -> Sequence[str]:
    from duckietown_world.world_duckietown import get_sampling_points, ChooseTime

    images = images or {}
    timeseries = timeseries or {}
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    fn_svg = os.path.join(output_dir, "drawing.svg")
    fn_html = os.path.join(output_dir, "drawing.html")

    timestamps = get_sampling_points(root)
    # logger.info(f'timestamps: {timestamps}')
    if len(timestamps) == 0:
        keyframes = SampledSequence[Timestamp]([0], [0])
    else:
        keyframes = SampledSequence[Timestamp](range(len(timestamps)), timestamps)
    # nkeyframes = len(keyframes)

    if area is None:
        areas = []
        all_keyframes = keyframes.values
        keyframes_for_extent = [all_keyframes[0], all_keyframes[-1]]
        for t in keyframes_for_extent:
            root_t = root.filter_all(ChooseTime(t))
            # print(i, root_t)
            rarea = get_extent_points(root_t)
            areas.append(rarea)
        area = reduce(RectangularArea.join, areas)

    logger.info("area: %s" % area)
    drawing, base = get_basic_upright2(fn_svg, area, pixel_size)
    # drawing.add(drawing.defs())
    gmg = drawing.g()
    base.add(gmg)

    static, dynamic = get_static_and_dynamic(root)

    t0 = keyframes.values[0]
    root_t0 = root.filter_all(ChooseTime(t0))
    g_static = drawing.g()
    g_static.attribs["class"] = "static"

    draw_recursive(drawing, root_t0, g_static, draw_list=static)
    base.add(g_static)

    obs_div = Tag(name="div")
    imagename2div = {}
    for name in images:
        imagename2div[name] = Tag(name="div")
        obs_div.append(imagename2div[name])

    # logger.debug('dynamic: %s' % dynamic)
    for i, t in keyframes:
        g_t = drawing.g()
        g_t.attribs["class"] = "keyframe keyframe%d" % i

        root_t = root.filter_all(ChooseTime(t))

        draw_recursive(drawing, root_t, g_t, draw_list=dynamic)
        base.add(g_t)

        for name, sequence in images.items():
            try:
                obs = sequence.at(t)
                updated = True
            except UndefinedAtTime:
                obs = sequence.at_or_previous(t)
                updated = False

            img = Tag(name="img")
            if isinstance(obs, bytes):
                data = obs
            else:
                data = obs.bytes_contents

            if height_of_stored_images is not None:
                data = get_resized_image(data, height_of_stored_images)
            img.attrs["src"] = data_encoded_for_src(data, "image/jpeg")
            # print('image %s %s: %.4fMB ' % (i, t, len(resized) / (1024 * 1024.0)))
            img.attrs["class"] = "keyframe keyframe%d" % i
            img.attrs["visualize"] = "hide"
            img.attrs["updated"] = int(updated)
            imagename2div[name].append(img)

    other = ""

    # language=html
    visualize_controls = """\
            <style>
            *[visualize_parts=false] {
                display: none;
            }
            </style>
        
            <p>
            <input id='checkbox-static' type="checkbox"  onclick="hideshow(this);" checked>static data</input>
            <input id='checkbox-textures' type="checkbox"  onclick="hideshow(this);" checked>textures</input>
            <input id='checkbox-axes' type="checkbox"  onclick="hideshow(this);">axes</input>
            <br/>
            <input id='checkbox-lane_segments' type="checkbox"  onclick="hideshow(this);">map lane 
            segments</input>
            (<input id='checkbox-lane_segments-control_points' type="checkbox"  onclick="hideshow(
            this);">control 
            points</input>)</p>
            </p>
           
            
            <p>
            <input id='checkbox-vehicles' type="checkbox"  onclick="hideshow(this);" checked>vehicles</input>
            <input id='checkbox-duckies' type="checkbox"  onclick="hideshow(this);" checked>duckies</input>
            <input id='checkbox-signs' type="checkbox"  onclick="hideshow(this);" checked>signs</input>
            <input id='checkbox-sign-papers' type="checkbox"  onclick="hideshow(this);" checked>signs 
            textures</input>
            <input id='checkbox-decorations' type="checkbox"  onclick="hideshow(this);" 
            checked>decorations</input>
          
            </p>
             <p>
            <input id='checkbox-current_lane' type="checkbox"  onclick="hideshow(this);">current lane</input>
            <input id='checkbox-anchors' type="checkbox"  onclick="hideshow(this);">anchor point</input>
            </p>
            <script>
                var checkboxValues = null; 
                name2selector = {
                    "checkbox-static": "g.static",
                    "checkbox-textures": "g.static .tile-textures",
                    "checkbox-axes": "g.axes",
                    "checkbox-lane_segments": "g.static .LaneSegment",
                    "checkbox-lane_segments-control_points": " .control-point",
                    "checkbox-current_lane": "g.keyframe .LaneSegment",
                    "checkbox-duckies": ".Duckie",
                    "checkbox-signs": ".Sign",
                    "checkbox-sign-papers": ".Sign .sign-paper",
                    "checkbox-vehicles": ".Vehicle",
                    "checkbox-decorations": ".Decoration",
                    'checkbox-anchors': '.Anchor',
                };
                function hideshow(element) {
                    console.log(element);
                    element_name = element.id;
                    console.log(element_name);
                    selector = name2selector[element_name];
                    checked = element.checked;
                    console.log(selector);
                    console.log(checked);
                    elements = document.querySelectorAll(selector);
                    elements.forEach(_ => _.setAttribute('visualize_parts', checked));
                    checkboxValues[element_name] = checked;
                    try {
                        localStorage.setItem("checkboxValues", JSON.stringify(checkboxValues));
                    } catch (error) {
                        console.log('cannot save preferences.');
                        console.log(error);
                    }
                }
                
                function init() {
                    for(var name in name2selector) {
                        console.log(name);
                        element = document.getElementById(name);
                        if(name in checkboxValues) {
                            element.checked = checkboxValues[name];
                        }
                        
                        hideshow(element);
                    } 
                }
                
                document.addEventListener("DOMContentLoaded", function(event) {
                    init();     
                });
                
                try {
                    checkboxValues =  JSON.parse(localStorage.getItem('checkboxValues')) || {};
                
                } catch (error) {
                    console.log('cannot load preferences.');
                    console.log(error);
                    checkboxValues = {}
                }
                
                init();
                console.log(checkboxValues);
            </script>
        """

    div_timeseries = str(make_tabs(timeseries))

    obs_div = str(obs_div)
    html = make_html_slider(
        drawing,
        keyframes,
        obs_div=obs_div,
        other=other,
        div_timeseries=div_timeseries,
        visualize_controls=visualize_controls,
    )
    with open(fn_html, "w") as f:
        f.write(html)

    # language=css
    style = """
        .sign-paper {
            display: none;
        }
        g.axes, .LaneSegment {
            display: none;
        }
         
    """
    drawing.defs.add(drawing.style(style))

    drawing.save(pretty=True)
    logger.info("Written SVG to %s" % fn_svg)
    logger.info("Written HTML to %s" % fn_html)

    return [fn_svg, fn_html]


def get_resized_image(bytes_content, width):
    pl = logging.getLogger("PIL")
    pl.setLevel(logging.ERROR)
    idata = BytesIO(bytes_content)
    with Image.open(idata) as _:
        image = _.convert("RGB")
    size = image.size
    height = int(size[1] * 1.0 / size[0] * width)
    image = image.resize((width, height))
    out = BytesIO()
    image.save(out, format="jpeg")

    return out.getvalue()


@dataclass
class TimeseriesPlot:
    title: str
    long_description: Optional[str]
    sequences: Dict

    # def __init__(self, title, long_description, sequences):
    #     check_isinstance(title, six.string_types)
    #     self.title = title
    #     self.long_description = long_description
    #     self.sequences = sequences

    def get_title(self):
        return self.title

    def get_long_description(self):
        return self.long_description


def make_tabs(timeseries):
    tabs = {}
    import plotly.offline as offline

    include_plotlyjs = True

    i = 0
    for name, tsp in timeseries.items():
        assert isinstance(tsp, TimeseriesPlot)

        div = Tag(name="div")
        table = Tag(name="table")
        table.attrs["style"] = "width: 100%"
        tr = Tag(name="tr")

        td = Tag(name="td")
        td.attrs["style"] = "width: 15em; min-height: 20em; vertical-align: top;"
        td.append(get_markdown(tsp.long_description or ""))
        tr.append(td)

        td = Tag(name="td")
        td.attrs["style"] = "width: calc(100%-16em); min-height: 20em; vertical-align: top;"

        import plotly.graph_objs as go
        import plotly.tools as tools

        scatters = []
        for name_sequence, sequence in tsp.sequences.items():
            assert isinstance(sequence, SampledSequence)

            trace = go.Scatter(
                x=sequence.timestamps, y=sequence.values, mode="lines+markers", name=name_sequence,
            )
            scatters.append(trace)

        layout = {"font": dict(size=10), "margin": dict(t=0)}

        n = len(scatters)
        if n == 0:
            p = Tag(name="p")
            p.append("No scatter plots available")
            td.append(p)
        else:

            fig = tools.make_subplots(rows=1, cols=n)
            fig.layout.update(layout)
            for j, scatter in enumerate(scatters):
                fig.append_trace(scatter, 1, j + 1)

            # include_plotlyjs = True if i == 0 else False

            res = offline.plot(fig, output_type="div", show_link=False, include_plotlyjs=include_plotlyjs, )
            include_plotlyjs = False
            td.append(bs(res))
            i += 1

        tr.append(td)
        table.append(tr)

        div.append(table)

        tabs[name] = Tab(title=tsp.title, content=div)

    return render_tabs(tabs)


class Tab:
    def __init__(self, title: str, content):
        self.title = title
        self.content = content


def render_tabs(tabs: Dict[str, Tab]) -> Tag:
    div_buttons = Tag(name="div")
    div_buttons.attrs["class"] = "tab"
    div_content = Tag(name="div")

    for i, (name, tab) in enumerate(tabs.items()):
        assert isinstance(tab, Tab), tab

        tid = "tab%s" % i
        button = Tag(name="button")
        button.attrs["class"] = "tablinks"
        button.attrs["onclick"] = "open_tab(event,'%s')" % tid
        button.append(tab.title)
        div_buttons.append(button)

        div_c = Tag(name="div")
        div_c.attrs["id"] = tid
        div_c.attrs["style"] = ""  # ''display: none; width:100%; height:100vh'

        div_c.attrs["class"] = "tabcontent"

        div_c.append(tab.content)

        div_content.append(div_c)

    script = Tag(name="script")
    # language=js
    js = """
function open_tab(evt, cityName) {
    // Declare all variables
    var i, tabcontent, tablinks;

    // Get all elements with class="tabcontent" and hide them
    tabcontent = document.getElementsByClassName("tabcontent");
    for (i = 0; i < tabcontent.length; i++) {
        tabcontent[i].style.display = "none";
    }

    // Get all elements with class="tablinks" and remove the class "active"
    tablinks = document.getElementsByClassName("tablinks");
    for (i = 0; i < tablinks.length; i++) {
        tablinks[i].className = tablinks[i].className.replace(" active", "");
    }

    // Show the current tab, and add an "active" class to the button that opened the tab
    document.getElementById(cityName).style.display = "block";
    document.getElementById(cityName).style.opacity = 1.0;
    evt.currentTarget.className += " active";
} 
    
    """
    script.append(js)

    style = Tag(name="style")
    # language=css
    style.append(
        """\
/* Style the tab */
.tab {
    overflow: hidden;
    border: 1px solid #ccc;
    background-color: #f1f1f1;
}

/* Style the buttons that are used to open the tab content */
.tab button {
    
    font-size: 80%;
    background-color: inherit;
    float: left;
    border: solid 0.5px gray;
    outline: none;
    cursor: pointer;
    /* padding: 14px 16px;*/
    transition: 0.3s;
}

.tab button + button {
    margin-left: 10px;
}

/* Change background color of buttons on hover */
.tab button:hover {
    background-color: #ddd;
}

/* Create an active/current tablink class */
.tab button.active {
    background-color: #ccc;
}

/* Style the tab content */
.tabcontent {
    /*display: none;*/
    opacity: 0;
    padding: 6px 12px;
    border: 1px solid #ccc;
    border-top: none;
    width: 100%;
}
    
    """
    )
    main = Tag(name="div")
    main.attrs["id"] = "tabs"
    main.append(style)
    main.append(script)
    main.append(div_buttons)
    main.append(div_content)
    return main


def make_html_slider(drawing, keyframes, obs_div, other, div_timeseries, visualize_controls):
    nkeyframes = len(keyframes.timestamps)

    # language=html
    controls_html = """\

<div id="slidecontainer">
<div id='fixedui'>
    Select time: <input autofocus type="range" min="0" max="%s" value="0" class="slider" id="time-range" 
    onchange="showVal(this.value)" oninput="showVal(this.value)"/>
    <span id="time-display"></span>
    </div>
</div>
<style type='text/css'>
    #slidecontainer {
    height: 3em;
    }
    #time-range {
    width: 50%%;
    }
    #fixedui { 
    position: fixed; 
    width: 100%%;
    height: 3em;
    background-color: white;
    }
    .keyframe[visualize="hide"] {
        display: none;
    }
    .keyframe[visualize="show"] {
        display: inherit;
    }
    td#obs {
        padding: 1em;
        vertical-align: top;
    }
    
    #observation_sequence {
        width: 320px;
    }
    td#obs img { width: 90%%;} 
</style>
<script type='text/javascript'>
    function showVal(newVal) {
        elements = document.querySelectorAll('.keyframe');
        elements.forEach(_ => _.setAttribute('visualize', 'hide'));
        elements_show = document.querySelectorAll('.keyframe' + newVal );  
        elements_show.forEach(_ => _.setAttribute('visualize', 'show'));
    }
    document.addEventListener("DOMContentLoaded", function(event) {
        showVal(0);
    });
</script>
""" % (
        nkeyframes - 1
    )

    if nkeyframes <= 1:
        controls_html += """
        <style>
        .slidecontainer {
        display: none;
        }
        </style>
        """

    controls = bs(controls_html)

    valbox = controls.find("span", id="time-display")
    assert valbox is not None
    for i, timestamp in keyframes:
        t = Tag(name="span")
        t.attrs["class"] = "keyframe keyframe%d" % i
        t.attrs["visualize"] = "hide"
        t.append("t = %.2f" % timestamp)

        valbox.append(t)

    from six import StringIO

    f = StringIO()
    drawing.write(f, pretty=True)
    drawing_svg = f.getvalue()
    f.close()
    # drawing_svg = drawing.tostring(pretty=True)
    # language=html
    doc = """\
<html lang='en'>
<head><title></title></head>
<body>
<style>
/*svg {{ background-color: #eee;}}*/
body {{
    font-family: system-ui, sans-serif;
}}
</style>
{controls}
<table>
<tr>
<td style="width: 640px; vertical-align:top;">
{drawing}
</td>
<td id="obs" >
{visualize_controls}
<div id="observation_sequence">
{obs_div}
</div>
</td>
</tr>
</table>

<script type='text/javascript'>
    showVal(0); 
</script>

{div_timeseries}
{other}
</body>
</html>
    """.format(
        controls=str(controls),
        drawing=drawing_svg,
        obs_div=obs_div,
        other=other,
        div_timeseries=div_timeseries,
        visualize_controls=visualize_controls,
    )
    return doc


def mime_from_fn(fn):
    if fn.endswith("png"):
        return "image/png"
    elif fn.endswith("jpg"):
        return "image/jpeg"
    else:
        raise ValueError(fn)


def data_encoded_for_src(data, mime):
    """ data =
        ext = png, jpg, ...

        returns "data: ... " sttring
    """
    encoded = base64.b64encode(data).decode()
    link = ("data:%s;base64," % mime) + encoded
    return link


def draw_axes(drawing, g, L=0.1, stroke_width=0.01, klass="axes"):
    g2 = drawing.g()
    g2.attribs["class"] = klass
    line = drawing.line(start=(0, 0), end=(L, 0), stroke_width=stroke_width, stroke="red")
    g2.add(line)

    line = drawing.line(start=(0, 0), end=(0, L), stroke_width=stroke_width, stroke="green")
    g2.add(line)

    g.add(g2)


@memoized_reset
def get_jpeg_bytes(fn):
    pl = logging.getLogger("PIL")
    pl.setLevel(logging.ERROR)

    with Image.open(fn) as _:
        image = _.convert("RGB")

    out = BytesIO()
    image.save(out, format="jpeg")
    return out.getvalue()


def bs(fragment: str):
    """ Returns the contents wrapped in an element called "fragment".
        Expects fragment as a str in utf-8 """

    s = u"<fragment>%s</fragment>" % fragment

    wire = s.encode("utf-8")
    parsed = BeautifulSoup(wire, "lxml")
    res = parsed.html.body.fragment
    assert res.name == "fragment"
    return res


def get_markdown(md: str) -> str:
    import markdown

    extensions = ["extra", "smarty"]
    html = markdown.markdown(md, extensions=extensions, output_format="html5")

    res = bs(html)
    return res



import random

import numpy as np

import geometry as geo
from duckietown_world import (
    iterate_by_class,
    LaneSegment,
    IterateByTestResult,
    Tile,
    logger,
    PlacedObject,
    LanePose,
    SE2Transform,
)

__all__ = ["sample_good_starting_pose"]


# @dataclass
# class SampledPose:
#     # pose
#     q: np.ndarray
#     # pose as an SE2Transform
#     qt: SE2Transform
#     # corresponding  tile
#     tile: Any


def sample_good_starting_pose(
    m: PlacedObject, only_straight: bool = True, along_lane: float = 0.2
) -> np.ndarray:
    """ Samples a good starting pose on a straight lane """
    choices = list(iterate_by_class(m, LaneSegment))

    if only_straight:
        choices = [_ for _ in choices if is_straight(_)]

    choice = random.choice(choices)
    ls: LaneSegment = choice.object

    lp: LanePose = ls.lane_pose(along_lane, 0.0, 0.0)
    rel: SE2Transform = ls.SE2Transform_from_lane_pose(lp)

    m1 = choice.transform_sequence.asmatrix2d().m
    m2 = rel.asmatrix2d().m
    g = np.dot(m1, m2)

    t, a, s = geo.translation_angle_scale_from_E2(g)
    g = geo.SE2_from_translation_angle(t, a)

    return g


def is_straight(choice: IterateByTestResult):
    segment: LaneSegment = choice.object
    lane_segment_is_straight = np.allclose(segment.get_lane_length(), 1.0)
    tiles = [_ for _ in choice.parents if isinstance(_, Tile)]
    assert len(tiles) == 1
    tile: Tile = tiles[0]
    tile_is_straight = tile.kind == "straight"
    return lane_segment_is_straight and tile_is_straight



# coding=utf-8

from duckietown_serialization_ds1 import Serializable

import geometry as geo
from .generic_kinematics import GenericKinematicsSE2
from .platform_dynamics import PlatformDynamicsFactory

__all__ = [
    "WheelVelocityCommands",
    "DifferentialDriveDynamicsParameters",
    "DifferentialDriveDynamics",
]


class WheelVelocityCommands(Serializable):
    """
        This represents the velocity commands for differential drive
        kinematics.

        These are expressed in rad/s.

        For both wheels, positive = moves forward.

    """

    left_wheel_angular_velocity: float
    right_wheel_angular_velocity: float

    def __init__(self, left_wheel_angular_velocity: float, right_wheel_angular_velocity: float):
        self.left_wheel_angular_velocity = left_wheel_angular_velocity
        self.right_wheel_angular_velocity = right_wheel_angular_velocity


class DifferentialDriveDynamicsParameters(PlatformDynamicsFactory, Serializable):
    """
        This class represents the parameters of the ideal differential drive dynamics.

        radius_left, radius_right: wheels radii
        wheel_distance: distance between two wheels

    """

    radius_left: float
    radius_right: float
    wheel_distance: float

    def __init__(self, radius_left, radius_right, wheel_distance):
        self.radius_left = radius_left
        self.radius_right = radius_right
        self.wheel_distance = wheel_distance

    def initialize(self, c0, t0=0, seed=None):
        return DifferentialDriveDynamics(self, c0, t0)


class DifferentialDriveDynamics(GenericKinematicsSE2):
    """
        This represents the state of differential drive.

        This is a particular case of GenericKinematicsSE2.

    """

    def __init__(self, parameters: DifferentialDriveDynamicsParameters, c0, t0):
        """
        :param parameters:  instance of DifferentialDriveDynamicsParameters
        :param c0: initial configuration
        :param t0: initial time
        """
        self.parameters = parameters
        GenericKinematicsSE2.__init__(self, c0, t0)

    def integrate(self, dt: float, commands: WheelVelocityCommands):
        """

        :param dt:
        :param commands: an instance of WheelVelocityCommands
        :return:
        """

        # Compute the linear velocity for the wheels
        # by multiplying radius times angular velocity
        v_r = self.parameters.radius_right * commands.right_wheel_angular_velocity
        v_l = self.parameters.radius_left * commands.left_wheel_angular_velocity

        # compute the linear, angular velocities for the platform
        # using the differential drive equations
        longitudinal = (v_r + v_l) * 0.5
        angular = (v_r - v_l) / self.parameters.wheel_distance
        lateral = 0.0

        linear = [longitudinal, lateral]

        # represent this as se(2)
        commands_se2 = geo.se2_from_linear_angular(linear, angular)

        # Call the "integrate" function of GenericKinematicsSE2
        s1 = GenericKinematicsSE2.integrate(self, dt, commands_se2)

        # new state
        c1 = s1.q0, s1.v0
        t1 = s1.t0
        return DifferentialDriveDynamics(self.parameters, c1, t1)
